/*
 * path_id_compat.c: compose persistent device path (compat version)
 *
 * Copyright (C) 2009 Kay Sievers <kay.sievers@vrfy.org>
 *
 * Logic based on Hannes Reinecke's shell script.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <dirent.h>

#include "device-util.h"
#include "dirent-util.h"
#include "fd-util.h"
#include "parse-util.h"
#include "sd-device.h"
#include "string-util.h"

#define PATH_SIZE       16384
#define SYSFS_PATH      "/sys"

static const char *compat_version_str = NULL;
static unsigned compat_version;

static int path_prepend(char **path, const char *fmt, ...) {
        va_list va;
        char *old;
        char *pre;
        int err;

        old = *path;

        va_start(va, fmt);
        err = vasprintf(&pre, fmt, va);
        va_end(va);
        if (err < 0)
                return err;

        if (old != NULL) {
                err = asprintf(path, "%s-%s", pre, old);
                if (err < 0)
                        return err;
                free(pre);
        } else {
                *path = pre;
        }

        free(old);
        return 0;
}

/*
** Linux only supports 32 bit luns.
** See drivers/scsi/scsi_scan.c::scsilun_to_int() for more details.
*/
static int format_lun_number(sd_device *dev, char **path) {
       unsigned long lun;
       const char *sysnum;

       (void) sd_device_get_sysnum(dev, &sysnum);

       lun = strtoul(sysnum, NULL, 10);
       if (lun < 256) /* address method 0, peripheral device addressing with bus id of zero */
               return path_prepend(path, "lun-%d", lun);

       /* handle all other lun addressing methods by using a variant of the original lun format */
       return path_prepend(path, "lun-0x%04x%04x00000000", (lun & 0xffff), (lun >> 16) & 0xffff);
}

static sd_device *skip_subsystem(sd_device *dev, const char *subsys) {
        for (sd_device *parent = dev; ; ) {
                if (!device_in_subsystem(parent, subsys))
                        break;

                dev = parent;
                if (sd_device_get_parent(dev, &parent) < 0)
                        break;
        }

        return dev;
}

static sd_device *handle_scsi_default(sd_device *parent, char **path) {
        sd_device *hostdev;
        int host, bus, target, lun;
        const char *sysname, *base, *pos;
        _cleanup_closedir_ DIR *dir = NULL;
        int basenum = -1;

        if (sd_device_get_parent_with_subsystem_devtype(parent, "scsi", "scsi_host", &hostdev) < 0)
                return NULL;

        if (sd_device_get_sysname(parent, &sysname) < 0)
                return NULL;
        if (sscanf(sysname, "%d:%d:%d:%d", &host, &bus, &target, &lun) != 4)
                return NULL;

        /* rebase host offset to get the local relative number */
        if (sd_device_get_syspath(hostdev, &base) < 0)
                return NULL;
        pos = strrchr(base, '/');
        if (!pos)
                return NULL;

        base = strndupa_safe(base, pos - base);
        dir = opendir(base);
        if (!dir)
                return NULL;

        FOREACH_DIRENT_ALL(dent, dir, break) {
                char *rest;
                int i;

                if (dent->d_name[0] == '.')
                        continue;
                if (dent->d_type != DT_DIR && dent->d_type != DT_LNK)
                        continue;
                if (strncmp(dent->d_name, "host", 4) != 0)
                        continue;
                i = strtoul(&dent->d_name[4], &rest, 10);
                if (rest[0] != '\0')
                        continue;
                if (basenum == -1 || i < basenum)
                        basenum = i;
        }
        if (basenum == -1)
                return NULL;
        host -= basenum;

        path_prepend(path, "scsi-%u:%u:%u:%u", host, bus, target, lun);
        return hostdev;
}

static sd_device *handle_ata(sd_device *parent, char **path) {
        sd_device *hostdev;
        int host, bus, target, lun;
        const char *name;

        if (sd_device_get_parent_with_subsystem_devtype(parent, "scsi", "scsi_host", &hostdev) < 0)
                return NULL;

        if (sd_device_get_sysname(parent, &name) < 0)
                return NULL;
        if (sscanf(name, "%d:%d:%d:%d", &host, &bus, &target, &lun) != 4)
                return NULL;

        path_prepend(path, "scsi-%u:%u:%u:%u", host, bus, target, lun);
        return hostdev;
}

static sd_device *handle_scsi_sas(sd_device *parent, char **path) {
        _cleanup_closedir_ DIR *dir = NULL;
        _cleanup_(sd_device_unrefp) sd_device *sasdev = NULL;
        sd_device *targetdev, *target_parent, *portdev;
        const char *sas_address, *syspath, *sysname;
        int tmp_phy_id, phy_id = 255;
        _cleanup_free_ char *lun = NULL;

        if (sd_device_get_parent_with_subsystem_devtype(parent, "scsi", "scsi_target", &targetdev) < 0)
                return NULL;

        if (sd_device_get_parent(targetdev, &target_parent) < 0)
                return NULL;

        if (sd_device_get_parent(target_parent, &portdev) < 0)
                return NULL;

        if (sd_device_get_syspath(portdev, &syspath) < 0)
                return NULL;
        dir = opendir(syspath);
        if (!dir)
                return NULL;

        FOREACH_DIRENT_ALL(dent, dir, break) {
                const char *name = dent->d_name;
                char *phy_id_str;

                if (dent->d_type != DT_LNK)
                        continue;

                if (strncmp(dent->d_name, "phy", 3) != 0)
                        continue;

                phy_id_str = strstr(name, ":");
                if (phy_id_str == NULL)
                        continue;

                phy_id_str++;

                tmp_phy_id = atoi(phy_id_str);
                if (tmp_phy_id >= 0 && tmp_phy_id < phy_id)
                        phy_id = tmp_phy_id;
        }

        if (phy_id == 255)
                return NULL;

        if (sd_device_get_sysname(target_parent, &sysname) < 0)
                return NULL;

        if (sd_device_new_from_subsystem_sysname(&sasdev, "sas_device", sysname) < 0)
                return NULL;

        if (sd_device_get_sysattr_value(sasdev, "sas_address", &sas_address) < 0)
                return NULL;

        if (format_lun_number(parent, &lun) < 0)
                return NULL;

        switch (compat_version) {
        case 1:
                path_prepend(path, "sas-phy%d-%s-%s", phy_id, sas_address, lun);
                break;
        case 2:
                path_prepend(path, "sas-%s-%s", sas_address, lun);
                break;
        }

        return parent;
}

static sd_device *handle_scsi(sd_device *parent, char **path) {
        const char *devtype;
        const char *name;
        int r;

        r = sd_device_get_devtype(parent, &devtype);
        if (r < 0 || strcmp(devtype, "scsi_device") != 0)
                return parent;

        /* lousy scsi sysfs does not have a "subsystem" for the transport */
        (void) sd_device_get_syspath(parent, &name);

        if (strstr(name, "/end_device-") != NULL)
                return handle_scsi_sas(parent, path);

        if (strstr(name, "/ata") != NULL)
                return handle_ata(parent, path);

        return handle_scsi_default(parent, path);
}

int main(int argc, char **argv) {
        static const struct option options[] = {
                { "compat", required_argument, NULL, 'V' },
        };
        _cleanup_(sd_device_unrefp) sd_device *dev = NULL;
        _cleanup_free_ char *path = NULL;
        char syspath[PATH_SIZE];
        int r = 1;

        for (;;) {
                int option;

                option = getopt_long(argc, argv, "v:", options, NULL);
                if (option == -1)
                        break;

                switch (option) {
                case 'V':
                        compat_version_str = optarg;
                        break;
                }
        }

        if (compat_version_str) {
                if (safe_atou(compat_version_str, &compat_version) < 0) {
                        fprintf(stderr, "--compat takes an integer.\n");
                        return 1;
                }
        }

        if (argv[optind] == NULL) {
                fprintf(stderr, "No device specified\n");
                return 2;
        }

        snprintf(syspath, PATH_SIZE, "%s%s", SYSFS_PATH, argv[optind]);
        if (sd_device_new_from_syspath(&dev, syspath) < 0) {
                fprintf(stderr, "unable to access '%s'\n", argv[optind]);
                return 3;
        }

        /* walk up the chain of devices and compose path */
        for (sd_device *parent = dev; parent; ) {
                const char *sysname;

                if (sd_device_get_sysname(parent, &sysname) < 0)
                        ;
                else if (device_in_subsystem(parent, "scsi"))
                        parent = handle_scsi(parent, &path);
                else if (device_in_subsystem(parent, "pci")) {
                        path_prepend(&path, "pci-%s", sysname);
                        parent = skip_subsystem(parent, "pci");
                }

                if (!parent)
                        break;
                if (sd_device_get_parent(parent, &parent) < 0) /* return -ENOENT when no more parent */
                        break;
        }

        if (path) {
                printf("ID_PATH_COMPAT%s=%s\n", strempty(compat_version_str), path);
                r = 0;
        }

        return r;
}
